#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ----------- FUNOES ---------------- */

/**********************************************************************
 *  Funcao imprimeMatriz
 * 
 *  Acao:       Imprime o conteudo de uma matriz em stdout
 *  Entrada:    ponteiro para matriz, ordem da matriz (linhas e colunas)
 *  Saida:      nenhuma
 *  Suposicoes: ponteiro para matriz deve ser nao nulo
 **********************************************************************
 */

void imprimeMatriz (double **matriz, int linhas, int colunas)
{
  int i, k;
  for (i=0; i < linhas; i++) {
    for (k=0; k < colunas; k++)
      printf("%5.2lf\t", matriz[i][k]);
    printf("\n");
  }
}

/**********************************************************************
 *  Funcao obtemOrdemMatriz
 * 
 *  Acao:       Obtem de um arquivo a ordem das matrizes que serao 
 *              lidas postriormente pela funcao "obtemMatrizes()"
 *  Entrada:    fp = descritor de arquivo
 *  Saida:      linhas e colunas = ordem das matrizes
 *  Retorno:    1 se leitura bem sucedida, 0 se chegou ao final do arquivo
 **********************************************************************
 */

int obtemOrdemMatriz (FILE *fp, int *linhas, int *colunas)
{
  return (fscanf (fp, "%d %d", linhas, colunas) != EOF);

}


/**********************************************************************
 *  Funcao obtemMatriz
 * 
 *  Acao:       Obtem do usuario matriz linhasXcolunas de um arquivo
 *  Entrada:    fp = descritor de arquivo
 *              linhas e colunas = ordem da matriz
 *  Retorno:    ponteiro para matriz.
 *              Se NULO, erro em alocacao de memoria
 *  Algoritmo:  espaco para matriz eh alocado com "calloc()"
 **********************************************************************
 */

double **obtemMatriz (FILE *fp, int linhas, int colunas)
{
  int i, k;
  double **matriz;

  if (matriz = (double **) calloc (linhas, sizeof(double *)))
    for (i = 0; matriz && i < linhas; i++) {
      if (matriz[i] = (double *) calloc (colunas, sizeof (double)))
	for (k = 0; k < colunas ; k++)
	  fscanf(fp, "%lf", &matriz[i][k]);
      else
	matriz = NULL;
    }

  return (matriz);
}

/**********************************************************************
 *  Funcao multMatrizes
 * 
 *  Acao:       Efetua multiplicacao entre duas matrizes
 *  Entrada:    mat_1 e mat_2 = matrizes a serem multiplicadas
 *              linres x collin = ordem de mat_1
 *              collin x colres = ordem de mat_2
 *  Retorno:    ponteiro para matriz com resultado (linres x colres)
 *              Se NULO, erro em alocacao de memoria
 *  Algoritmo:  espaco para matriz com resultado eh alocada com "calloc()"
 **********************************************************************
 */

double **multMatrizes (double **mat_1, double **mat_2, 
		     int linres, int collin, int colres)
{
  int i, j, p;
  double **prod; /* Resultado da multiplicacao */

  /* Cria e inicia em 0 (zero) a matriz que receber o resultado 
   * da multiplicaao entre mat_1 e mat_2
   */
  if (prod = (double **) calloc (linres, sizeof(double *)))
    for (i=0; prod && i < linres; i++)
      if (!(prod[i] = (double*) calloc (colres, sizeof(double))))
	prod = NULL;
    
  /* Efetua a multiplicaao */
  if (prod)
    for (i=0; i < linres; i++)
      for (j=0; j < colres; j++)
	for (p=0; p < collin; p++)
	  prod[i][j] += mat_1[i][p] * mat_2[p][j];
  
  return (prod);
}

/**********************************************************************
 *  Funcao salvaResult
 * 
 *  Acao:       Salva matriz e sua ordem em arquivo
 *  Entrada:    file = nome do arquivo de entrada
 *              matriz, linhas e colunas = matriz e sua ordem
 *  Retorno:    1 se salvou com sucesso
 *              0 caso contrario
 **********************************************************************
 */

int salvaResult (char * file, double **matriz,
		  int linhas, int colunas)
{
  int i, k, ok;
  FILE *fpsai;
  char *arqsai;

  /* Gera nome do arquivo de saida = "res-" + nome arq. entrada */
  arqsai = strdup("res-");
  arqsai = realloc (arqsai, strlen(arqsai) + strlen(file) + 1);
  arqsai = strcat (arqsai, file);

  /* Abre arquivo de saida para anexar e grava resultado */
  if (ok = (fpsai = fopen (arqsai, "a")) != NULL) {
    fprintf(fpsai, "%d %d\n", linhas, colunas);
    for (i=0; i < linhas; i++) {
      for (k=0; k < colunas; k++)
	fprintf(fpsai, "%5.2lf ", matriz[i][k]);
      fprintf(fpsai, "\n");
    }
    
    fclose (fpsai);
  }
  else
    fprintf (stderr, "Erro na abertura do arq. \"%s\"\n", arqsai);

  return (ok);
}

/**********************************************
 * Programa principal
 * Forma de uso: prog [arquivo]
 *               se "arquivo" nao eh fornecido,
 *               assume "matriz_01.dat"
 **********************************************
 */

main (int argc, char *argv[]) 
{
  char *arqent = "matriz_01.dat";
  FILE *fpent;

  double **matriz,
         **mult;
  int i,k,
      linhas, colunas;

  if (argc == 2)
    arqent = argv[1];
  
  if (!(fpent = fopen (arqent, "r")))
    fprintf (stderr, "Erro na abertura do arq. \"%s\"\n", arqent);
  else {
    while (obtemOrdemMatriz(fpent, &linhas, &colunas))
    {
      if (matriz = obtemMatriz (fpent, linhas, colunas)) {
	if (mult = obtemMatriz (fpent, colunas, linhas)) {
	  mult = multMatrizes(matriz, mult, linhas, colunas, linhas);
	  if (!salvaResult(arqent, mult, linhas, linhas))
	    exit (2);
	}
      }

      if (! matriz || ! mult) {
	fprintf (stderr, "Erro na alocaao de espao para matrizes\n");
	exit(1);
      }
    }
    fclose (fpent);
  }
}
